! Copyright (C) 2015 Arne Luechow
!
! SPDX-License-Identifier: GPL-3.0-or-later

module sphericalIntGrids_m

   use kinds_m, only: r8
   implicit none
   public
   private :: i,is2,is3,lam,rho

   integer, parameter :: NGridPointsArray(8) = [4, 6, 12, 14, 18, 26, 38, 50]
   integer, parameter :: MaxRulePoints=168
   integer, parameter :: RuleIdx(9) = [0,4,10,22,36,54,80,118,168]
   integer            :: i

   real(r8), parameter :: is2 = 0.707106781186547462d0   ! 1 / sqrt(2)
   real(r8), parameter :: is3 = 0.577350269189625842d0   ! 1 / sqrt(3)
   real(r8), parameter :: lam = sqrt((5d0 - sqrt(5d0))/10d0)
   real(r8), parameter :: rho = sqrt((5d0 + sqrt(5d0))/10d0)

   real(r8), parameter :: AllGridPoints(3,MaxRulePoints) = reshape(  [ &
      ! tetrahedral rule
       is3,  is3,  is3,    &
       is3, -is3, -is3,    &
      -is3,  is3, -is3,    &
      -is3, -is3,  is3,    &
      ! octahedral rule
       1d0,  0d0,  0d0,    &
      -1d0,  0d0,  0d0,    &
       0d0,  1d0,  0d0,    &
       0d0, -1d0,  0d0,    &
       0d0,  0d0,  1d0,    &
       0d0,  0d0, -1d0,    &
      ! icosahedral rule
       0d0,  lam,  rho,    &
       0d0, -lam,  rho,    &
       0d0,  lam, -rho,    &
       0d0, -lam, -rho,    &
       rho,  0d0,  lam,    &
       rho,  0d0, -lam,    &
      -rho,  0d0,  lam,    &
      -rho,  0d0, -lam,    &
       lam,  rho,  0d0,    &
      -lam,  rho,  0d0,    &
       lam, -rho,  0d0,    &
      -lam, -rho,  0d0,    &
      ! Lebedev rule 14 (octahedron+cube)
       1d0,  0d0,  0d0,    &
      -1d0,  0d0,  0d0,    &
       0d0,  1d0,  0d0,    &
       0d0, -1d0,  0d0,    &
       0d0,  0d0,  1d0,    &
       0d0,  0d0, -1d0,    &
       is3,  is3,  is3,    &
      -is3,  is3,  is3,    &
       is3, -is3,  is3,    &
       is3,  is3, -is3,    &
      -is3, -is3,  is3,    &
      -is3,  is3, -is3,    &
       is3, -is3, -is3,    &
      -is3, -is3, -is3,    &
      ! 18 point rule (octahedron+points above edges)
       1d0,  0d0,  0d0,    &
      -1d0,  0d0,  0d0,    &
       0d0,  1d0,  0d0,    &
       0d0, -1d0,  0d0,    &
       0d0,  0d0,  1d0,    &
       0d0,  0d0, -1d0,    &
       is2,  is2,  0d0,    &
       is2, -is2,  0d0,    &
      -is2,  is2,  0d0,    &
      -is2, -is2,  0d0,    &
       is2,  0d0,  is2,    &
       is2,  0d0, -is2,    &
      -is2,  0d0,  is2,    &
      -is2,  0d0, -is2,    &
       0d0,  is2,  is2,    &
       0d0,  is2, -is2,    &
       0d0, -is2,  is2,    &
       0d0, -is2, -is2,    &
      ! Lebedev rule 26 point (18 point + cube)
       1d0,  0d0,  0d0,    &
      -1d0,  0d0,  0d0,    &
       0d0,  1d0,  0d0,    &
       0d0, -1d0,  0d0,    &
       0d0,  0d0,  1d0,    &
       0d0,  0d0, -1d0,    &
       is2,  is2,  0d0,    &
       is2, -is2,  0d0,    &
      -is2,  is2,  0d0,    &
      -is2, -is2,  0d0,    &
       is2,  0d0,  is2,    &
       is2,  0d0, -is2,    &
      -is2,  0d0,  is2,    &
      -is2,  0d0, -is2,    &
       0d0,  is2,  is2,    &
       0d0,  is2, -is2,    &
       0d0, -is2,  is2,    &
       0d0, -is2, -is2,    &
       is3,  is3,  is3,    &
      -is3,  is3,  is3,    &
       is3, -is3,  is3,    &
       is3,  is3, -is3,    &
      -is3, -is3,  is3,    &
      -is3,  is3, -is3,    &
       is3, -is3, -is3,    &
      -is3, -is3, -is3,    &
      ! Lebedev grid n=38 (from matlab script getLebedevSphere (van Wuellen))
  1.00000000000000d0,    0.00000000000000d0,    0.00000000000000d0,   &
 -1.00000000000000d0,    0.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,    1.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,   -1.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,    0.00000000000000d0,    1.00000000000000d0,   &
  0.00000000000000d0,    0.00000000000000d0,   -1.00000000000000d0,   &
  0.57735026918963d0,    0.57735026918963d0,    0.57735026918963d0,   &
 -0.57735026918963d0,    0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,   -0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,    0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,   -0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,   -0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,    0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,   -0.57735026918963d0,   -0.57735026918963d0,   &
  0.45970084338098d0,    0.88807383397712d0,    0.00000000000000d0,   &
 -0.45970084338098d0,    0.88807383397712d0,    0.00000000000000d0,   &
  0.45970084338098d0,   -0.88807383397712d0,    0.00000000000000d0,   &
 -0.45970084338098d0,   -0.88807383397712d0,    0.00000000000000d0,   &
  0.88807383397712d0,    0.45970084338098d0,    0.00000000000000d0,   &
 -0.88807383397712d0,    0.45970084338098d0,    0.00000000000000d0,   &
  0.88807383397712d0,   -0.45970084338098d0,    0.00000000000000d0,   &
 -0.88807383397712d0,   -0.45970084338098d0,    0.00000000000000d0,   &
  0.45970084338098d0,    0.00000000000000d0,    0.88807383397712d0,   &
 -0.45970084338098d0,    0.00000000000000d0,    0.88807383397712d0,   &
  0.45970084338098d0,    0.00000000000000d0,   -0.88807383397712d0,   &
 -0.45970084338098d0,    0.00000000000000d0,   -0.88807383397712d0,   &
  0.88807383397712d0,    0.00000000000000d0,    0.45970084338098d0,   &
 -0.88807383397712d0,    0.00000000000000d0,    0.45970084338098d0,   &
  0.88807383397712d0,    0.00000000000000d0,   -0.45970084338098d0,   &
 -0.88807383397712d0,    0.00000000000000d0,   -0.45970084338098d0,   &
  0.00000000000000d0,    0.45970084338098d0,    0.88807383397712d0,   &
  0.00000000000000d0,   -0.45970084338098d0,    0.88807383397712d0,   &
  0.00000000000000d0,    0.45970084338098d0,   -0.88807383397712d0,   &
  0.00000000000000d0,   -0.45970084338098d0,   -0.88807383397712d0,   &
  0.00000000000000d0,    0.88807383397712d0,    0.45970084338098d0,   &
  0.00000000000000d0,   -0.88807383397712d0,    0.45970084338098d0,   &
  0.00000000000000d0,    0.88807383397712d0,   -0.45970084338098d0,   &
  0.00000000000000d0,   -0.88807383397712d0,   -0.45970084338098d0,   &
    ! Lebedev grid n=50
  1.00000000000000d0,    0.00000000000000d0,    0.00000000000000d0,   &
 -1.00000000000000d0,    0.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,    1.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,   -1.00000000000000d0,    0.00000000000000d0,   &
  0.00000000000000d0,    0.00000000000000d0,    1.00000000000000d0,   &
  0.00000000000000d0,    0.00000000000000d0,   -1.00000000000000d0,   &
  0.00000000000000d0,    0.70710678118655d0,    0.70710678118655d0,   &
  0.00000000000000d0,   -0.70710678118655d0,    0.70710678118655d0,   &
  0.00000000000000d0,    0.70710678118655d0,   -0.70710678118655d0,   &
  0.00000000000000d0,   -0.70710678118655d0,   -0.70710678118655d0,   &
  0.70710678118655d0,    0.00000000000000d0,    0.70710678118655d0,   &
  0.70710678118655d0,    0.00000000000000d0,   -0.70710678118655d0,   &
 -0.70710678118655d0,    0.00000000000000d0,    0.70710678118655d0,   &
 -0.70710678118655d0,    0.00000000000000d0,   -0.70710678118655d0,   &
  0.70710678118655d0,    0.70710678118655d0,    0.00000000000000d0,   &
 -0.70710678118655d0,    0.70710678118655d0,    0.00000000000000d0,   &
  0.70710678118655d0,   -0.70710678118655d0,    0.00000000000000d0,   &
 -0.70710678118655d0,   -0.70710678118655d0,    0.00000000000000d0,   &
  0.57735026918963d0,    0.57735026918963d0,    0.57735026918963d0,   &
 -0.57735026918963d0,    0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,   -0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,    0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,   -0.57735026918963d0,    0.57735026918963d0,   &
  0.57735026918963d0,   -0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,    0.57735026918963d0,   -0.57735026918963d0,   &
 -0.57735026918963d0,   -0.57735026918963d0,   -0.57735026918963d0,   &
  0.30151134457776d0,    0.30151134457776d0,    0.90453403373329d0,   &
 -0.30151134457776d0,    0.30151134457776d0,    0.90453403373329d0,   &
  0.30151134457776d0,   -0.30151134457776d0,    0.90453403373329d0,   &
  0.30151134457776d0,    0.30151134457776d0,   -0.90453403373329d0,   &
 -0.30151134457776d0,   -0.30151134457776d0,    0.90453403373329d0,   &
 -0.30151134457776d0,    0.30151134457776d0,   -0.90453403373329d0,   &
  0.30151134457776d0,   -0.30151134457776d0,   -0.90453403373329d0,   &
 -0.30151134457776d0,   -0.30151134457776d0,   -0.90453403373329d0,   &
 -0.30151134457776d0,    0.90453403373329d0,    0.30151134457776d0,   &
  0.30151134457776d0,   -0.90453403373329d0,    0.30151134457776d0,   &
  0.30151134457776d0,    0.90453403373329d0,   -0.30151134457776d0,   &
 -0.30151134457776d0,   -0.90453403373329d0,    0.30151134457776d0,   &
 -0.30151134457776d0,    0.90453403373329d0,   -0.30151134457776d0,   &
  0.30151134457776d0,   -0.90453403373329d0,   -0.30151134457776d0,   &
 -0.30151134457776d0,   -0.90453403373329d0,   -0.30151134457776d0,   &
  0.30151134457776d0,    0.90453403373329d0,    0.30151134457776d0,   &
  0.90453403373329d0,    0.30151134457776d0,    0.30151134457776d0,   &
 -0.90453403373329d0,    0.30151134457776d0,    0.30151134457776d0,   &
  0.90453403373329d0,   -0.30151134457776d0,    0.30151134457776d0,   &
  0.90453403373329d0,    0.30151134457776d0,   -0.30151134457776d0,   &
 -0.90453403373329d0,   -0.30151134457776d0,    0.30151134457776d0,   &
 -0.90453403373329d0,    0.30151134457776d0,   -0.30151134457776d0,   &
  0.90453403373329d0,   -0.30151134457776d0,   -0.30151134457776d0,   &
 -0.90453403373329d0,   -0.30151134457776d0,   -0.30151134457776d0    &
  ], &
   [ 3, MaxRulePoints ]  &
)

   real(r8), parameter :: AllWeights(MaxRulePoints) = [   &
      (0.25d0,i=1,4),                 &  ! 4 point: 1/4
      (0.1666666666666667d0,i=1,6),   &  ! 6 point: 1/6
      (0.0833333333333333d0,i=1,12),  &  ! 12 point: 1/12
      (0.0666666666666666d0,i=1,6),   &  ! 14 point  
      (0.075d0,i=1,8),                &
      (0.0333333333333333d0,i=1,6),   &  ! 18 point rule
      (0.0666666666666666d0,i=1,12),  &
      (0.476190476190476164d-01,i=1,6),  &  ! 26 point rule
      (0.380952380952380987d-01,i=1,12), &
      (0.321428571428571397d-01,i=1,8),  &
      (0.00952380952381d0,i=1,6),     &     ! 38 point rule
      (0.03214285714286d0,i=1,8),     &
      (0.02857142857143d0,i=1,24),    & 
      (0.01269841269841d0,i=1,6),     &     ! 50 point rule
      (0.02257495590829d0,i=1,12),    & 
      (0.02109375000000d0,i=1,8),     &
      (0.02017333553792d0,i=1,24)     & 
   ]

end module sphericalIntGrids_m